/** 
Copyright (c) 2003 Helsinki Institute for Information Technology

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included 
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
*/

import com.hp.hpl.mesa.rdf.jena.mem.ModelMem;
import com.hp.hpl.mesa.rdf.jena.model.*;
import com.hp.hpl.mesa.rdf.jena.vocabulary.*;
//import com.hp.hpl.mesa.rdf.jena.query.*;
import com.hp.hpl.mesa.rdf.jena.common.*;
import com.hp.hpl.jena.n3.*;
import com.hp.hpl.jena.daml.*;
import com.hp.hpl.jena.daml.common.*;

import org.apache.xerces.parsers.DOMParser;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import java.io.*;
import java.util.*;



/** Terminaattori on museoesineiden automaattiseen rdf-annotointiin liittyv ohjelma, 
    jolla on tarkoitus tehd esitytettyj termikortteja. **/

public class Terminaattori{

    private Document document;

    //sisiset tietorakenteet 
    private Model globalTermsModel;   //olemassaoleva globaali termitiedosto luetaan thn
    private Model lokalTermsModel;   //olemassaoleva lokaali termitiedosto luetaan thn
    private Hashtable existingTerms; //olemassa olevat termit parsitaan termsModelista tnne
    private Hashtable newTerms; //uudet termit tallennetaan tnne aluksi
    private Properties config;    //sisinen tietorakenne configuraatio-tiedoille
    private String[] wantedElements; //tnne tallennetaan tieto niiden kenttien nimist,
                                     //joista halutaan kert termit
    private String nimiavaruus;

    //luku ja kirjoitus
    private FileWriter writer;
    private FileReader reader;
    
    
    /** konstruktori 
      */
    public Terminaattori (String xmlFileName, String configFileName) {
	
	//Luetaan config-tiedosto Properties-olioon
	try{
	    this.config = new Properties();
	    this.config.load(new FileInputStream(configFileName));
	} catch(Exception e){
	    System.out.println("Configuration.txt-tiedoston lukemisessa ongelma: ");
	    e.printStackTrace();
	}	

	//Tehdn tietorakenne johon merkitn, mitk elementit halutaan xml-kortista.
	//Elementtien nimet kerrottu configuration.txt-tiedostossa
	StringTokenizer elements = new StringTokenizer(this.config.getProperty("XMLkentat").trim(),",");
	if(elements != null || elements.countTokens() >0) {
	    wantedElements = new String[elements.countTokens()];
	    int counter = 0;
	    while (elements.hasMoreTokens()) {
		wantedElements[counter] = elements.nextToken().trim();			
		++counter;
	    }
	} else wantedElements = new String[0];

	//Luodaan xerces DOM Parser xml-korteista
	DOMParser parser = new DOMParser();	
	try {
	    //parsittavan xml-tiedoston nimi on kerrottu configuration.txt-tiedostossa
	    parser.parse(xmlFileName);    
	    this.document = parser.getDocument();
	} catch (SAXException e) {
	    System.out.println("Ongelma!: XML-tiedosto ei validoitunut: ");
	    System.err.println (e);
	} catch (IOException e) {
	    System.out.println("Ongelma!: XML-tiedoston lukemisessa io-ongelma: ");
	    System.err.println (e);
	}
	NodeList cards = this.getCards(); //oma metodi, esiintyy alempana

	//Luodaan Model termeist
	String globaaliTermitiedosto = this.config.getProperty("globaaliTermitiedosto").trim();
	String lokaaliTermitiedosto = this.config.getProperty("lokaaliTermitiedosto").trim();
	this.nimiavaruus = this.config.getProperty("nimiavaruus").trim();
	try {
	    
	    this.reader = new FileReader(globaaliTermitiedosto);
	    this.globalTermsModel = new ModelMem().read(this.reader,
						  nimiavaruus);
	    this.reader = new FileReader(lokaaliTermitiedosto);
	    this.lokalTermsModel = new ModelMem().read(this.reader,
						       nimiavaruus);
	    this.reader = null;
	} catch (FileNotFoundException e) {
	    System.out.println("Termitiedostoja ei lytynyt: ");
	    System.err.println (e);
	} catch (RDFException e) {
	    System.out.println("Termitiedostot ei valideja: ");
	    System.err.println (e);
	}	
	existingTerms = this.getExistingTerms();
	newTerms = new Hashtable();
        System.out.println("  Olemassa olevat termit on luettu. Niiden lukumr on "+existingTerms.size());
    }


    /** Metodi parsii rdf:st kaikki yksikt ja monikot. 
      */
    private Hashtable getExistingTerms() {
	Hashtable table = new Hashtable();
	try {
	    this.getExistingTerms(table, nimiavaruus+"singular", 
				  globalTermsModel);
	    this.getExistingTerms(table, nimiavaruus+"plural", 
				  globalTermsModel);
	    this.getExistingTerms(table, nimiavaruus+"singular", 
				  lokalTermsModel);
	    this.getExistingTerms(table, nimiavaruus+"plural", 
				  lokalTermsModel);
	} catch (RDFException e) {
	    System.out.println("Termitiedostossa ongelmia: ");
	    System.err.println (e);
	}
	return table;
    }

    /** Metodi auttaa getExistingTerms -metodia.
      */
    private void getExistingTerms(Hashtable table, String property, 
				  Model model)  throws RDFException{
	NodeIterator niter = model.
	    listObjectsOfProperty(new PropertyImpl(property));
	while(niter.hasNext()) {
	    String obj = niter.next().toString();
	    obj = obj.trim().toLowerCase();
	    table.put(obj, "0");
	}	
    }


    /** Metodi hakee Nodelistiin kaikki esinekortit XML:st, 
      * eli yksi kortti sislt aina yhden esineen tiedot. 
      */
    private NodeList getCards(){
	return this.document.getElementsByTagName("esine"); //esine!  
	//HUOM! Tmn korjattava, jos kyseess onkin esim. taulut tai muinaismuistot
    }

    
    /** Tm metodi saa parametrina listan XML-tiedoston 
      * esinekorteista. 
      * -> Kortit kydn lpi yksi kerrallaan.
      * -> Korteista erotetaan configuration.txt-tiedostossa 
      *    kerrottujen kenttien tiedot. 
      * -> Tietoja vertaillaan olemassa oleviin globaaleihin ja 
      *    lokaaleihin termeihin
      *    -> Jos termi lytyy sielt, sille ei tehd mitn.
      *    -> Jos termi ei viel lydy, se tallennetaan listaan ja
      *       sen esiintymien lukumr lasketaan (vhitellen)
      */    
    private void findTermCandidatesFromXML(NodeList list){
	
	//Kortit kydn lpi yksi kerrallaan
	for (int i = 0;i < list.getLength();i++){
	    Node node = list.item(i);
	    //System.out.println("------------------Uusi kortti------------------");

	    //Kydn lpi kortin kaikki elementit
	    NodeList cardElements = ((Element)node).getElementsByTagName("*");
	    for (int j=0; j<cardElements.getLength(); j++){  
		Node cardElementnode = cardElements.item(j);
		
		try {
		    //Korteista erotetaan configuration.txt-tiedostossa kerrottujen kenttien tiedot.
		    for (int k=0; k<wantedElements.length; ++k) {
		    if(wantedElements[k] != null && 
		       cardElementnode.getNodeName().compareTo(wantedElements[k])==0) { 
			//System.out.println("Token lydetty: "+wantedElements[k]+
			//                   ":"+cardElementnode.getNodeValue());

		       	    String infostring = "";
			    infostring = this.recurseNodeChildren(cardElementnode);
			    this.checkNewCandidates(infostring, wantedElements[k]);
			    
		    }
		    }
		    
		} catch(Exception e) {
		    System.out.print("Tkkii metodissa createListOfNewTerms: ");
		    e.printStackTrace();
		}
	    }
	}
    }
	

    /** Metodi ky lpi kaikki parametrina annetun noden lapset ja sisllt 
     * ja erottaa niist termikandidaatit.
     */
    private String recurseNodeChildren(Node node) {
	try{
	    if(node==null) return "";
	    else if(node.getNodeName().compareTo("#text")==0) {
		//System.out.println("DEBUG: recurse & text: "+node.getNodeValue());
		return node.getNodeValue().trim().toLowerCase();
	    }
	    else {
		//System.out.println("DEBUG: recurse & not text: parent: "+node.getNodeName());
		NodeList datanodes = ((Element)node).getChildNodes(); 
		String palautettava = "";
		int lkm = datanodes.getLength();
		//System.out.println("DEBUG: recurse & children: "+lkm);
		for(int i=0; i<lkm; i++){
		    Node child = datanodes.item(i);
		    //System.out.println("DEBUG: recurse & not text: "+i+":"+node.getNodeName()+":"+node.getNodeValue());
		    palautettava = palautettava + recurseNodeChildren(child) + ", ";
		}
		return palautettava;
	    }
	} catch(Exception e){
	    System.out.println("DEBUG: recurse & error: "+e);
	    return "";
	}
    }
		
    /** Metodi tutkii parametrina annetusta merkkijonosta kaikki termikandidaatit.
      * Jos termikandidaatti on jo olemassa oleva termi, sit ei list uusien termien listaan.
      * Jos termikandidaattia ei lydy vanhoista, se listn listaan.
      */
    private void checkNewCandidates(String candidateList, String elementname) {
	//Pilkotaan saatu arvo paloiksi vlimerkkien kohdalta
	StringTokenizer tok = new StringTokenizer(candidateList, ",.;:()");
	
	//Pilkottuja paloja vertaillaan olemassa oleviin globaaleihin ja 
	//lokaaleihin termeihin
	while(tok.hasMoreTokens()){
	    String termCandidate = tok.nextToken().trim().toLowerCase();
	    if(!existingTerms.containsKey(termCandidate)) {		
		//Jos termi ei viel lydy, se tallennetaan listaan ja
		//sen esiintymien lukumr lasketaan.
		//Mukana talletetaan tieto, mist elementist kyseinen kandidaatti lytyi
		String[] value = ((String[]) newTerms.get(termCandidate));
		//System.out.println("!!Infostring="+infostring+": "+value);  //DEBUG
		int num = 1;
		if(value!=null) {
		    if(value[0]!=null){
			num = new Integer(value[0]).intValue();
			++num;
		    }
		    if(value[1]!=null && value[1].length()>1) {
			elementname = value[1] +", "+elementname;
		    }
		} else {
		    value = new String[2];
		}

		value[0] = ""+num;
		value[1] = elementname;		
		newTerms.put(termCandidate, value);
	    }
	}
    }


    /** Metodi tekee uusista termeist Modelin ja
      * tulostaa sen. 
      */ 
    private void writeNewTermsRDF(){
	try{
	    if(newTerms != null) {
		Enumeration terms = newTerms.keys();
		while(terms.hasMoreElements()) {
		    String newTerm_tmp = (String) terms.nextElement();
		    String uri = this.makeUri(newTerm_tmp);
		    System.out.println(" Tehdn termi \""+newTerm_tmp+"\", jonka uriksi tulee \""+uri+"\"");
		    lokalTermsModel.add(new ResourceImpl(nimiavaruus+uri, lokalTermsModel ), 
				   new PropertyImpl("http://www.w3.org/1999/02/22-rdf-syntax-ns#type", lokalTermsModel ), 
				   new ResourceImpl(nimiavaruus+"tuntematon", lokalTermsModel));

		    if(newTerm_tmp.endsWith("t")) {
			System.out.println("    Luodaan termille monikko.");
			lokalTermsModel.add(new ResourceImpl(nimiavaruus+uri,lokalTermsModel ), 
				       new PropertyImpl(nimiavaruus+"plural",lokalTermsModel ), 
				       new LiteralImpl(newTerm_tmp)); 
		    } else {
			System.out.println("    Luodaan termille yksikk.");
			lokalTermsModel.add(new ResourceImpl(nimiavaruus+uri,lokalTermsModel ), 
				       new PropertyImpl(nimiavaruus+"singular",lokalTermsModel ), 
				       new LiteralImpl(newTerm_tmp));
		    }
		    lokalTermsModel.add(new ResourceImpl(nimiavaruus+uri,lokalTermsModel ), 
				   new PropertyImpl(nimiavaruus+"usage",lokalTermsModel ), 
				   new LiteralImpl("in_use"));

		    String selite = this.config.getProperty("selite").trim();		    
		    //for (int k=0; k<wantedElements.length; ++k)
		    //	if(wantedElements[k] != null)
		    //	    selite = selite + wantedElements[k]+ " ";
		    String[] value = ((String[]) newTerms.get(newTerm_tmp));
		    if(value!=null && value[1]!=null && value[1].length()>1) 
			    selite = selite + "; " + value[1];
		    
		    System.out.println("    Luodaan selite \""+selite+"\"");
		    lokalTermsModel.add(new ResourceImpl(nimiavaruus+uri,lokalTermsModel ), 
					new PropertyImpl(nimiavaruus+"source",lokalTermsModel ), 
					new LiteralImpl(selite));
		    
		    if(value!=null && value[0]!=null && value[0].length()>0) 
			lokalTermsModel.add(new ResourceImpl(nimiavaruus+uri,lokalTermsModel ), 
					    new PropertyImpl(nimiavaruus+"comments",lokalTermsModel ), 
					    new LiteralImpl("Uusi termi lytyi "+value[0]+" esineest."));
		}
	    }

	    System.out.println("");
	    System.out.println("Uusia termej oli yhteens "+newTerms.size()+" kpl.");
	    BufferedWriter rdfWriter = new BufferedWriter(
				       new FileWriter(this.config.getProperty("lokaaliTermitiedosto")));
 	    lokalTermsModel.write(rdfWriter, "RDF/XML", null);
            lokalTermsModel = null;
	}
	catch(Exception e){
	    System.out.print("Uusia termej ei saatu muunnettua RDFksi: ");
	    e.printStackTrace();
	}
    }
    
    private String makeUri(String word) {
	word = word.trim();
	word = word.toLowerCase();
        word = word.replaceAll("%", "%25"); //% = 37  (10-jrj) = 0010 0101 (2-jrj) = 25 (16-jrj) 
	word = word.replaceAll("", "%E4"); // = 228 (10-jrj) = 1110 0100 (2-jrj) = E4 (16-jrj) 
        word = word.replaceAll("", "%E5"); // = 229 (10-jrj) = 1110 0101 (2-jrj) = E5 (16-jrj) 
        word = word.replaceAll("", "%F6"); // = 246 (10-jrj) = 1111 0110 (2-jrj) = F6 (16-jrj) 
        word = word.replaceAll("", "%E7"); // = 233 (10-jrj) = 1110 1001 (2-jrj) = E9 (16-jrj) 
        word = word.replaceAll("", "%E8"); // = 232 (10-jrj) = 1110 1000 (2-jrj) = E8 (16-jrj) 
	word = word.replaceAll("", "%FC"); // = 252 (10-jrj) = 1111 1100 (2-jrj) = FC (16-jrj) 
        word = word.replaceAll("", "%EA"); // = 234 (10-jrj) = 1110 1010 (2-jrj) = EA (16-jrj) 
        word = word.replace(' ', '_');
        word = word.replace(',', '_');
        word = word.replace(':', '_');
        word = word.replace('.', '_');
        return word;
    }


    /** Metodi printtaa uudet termit ja niiden lukumrt 
      * testitarkoitukseen. 
      */
    public void printNewTerms() {
	if(newTerms == null) {
	    System.out.println("Ei ole termej printattavaksi.");
	    return; 
	}
	Enumeration terms = newTerms.keys();
	System.out.println("Uusia termej on "+newTerms.size()+" kpl");
	while(terms.hasMoreElements()) {
	    String tmp = (String) terms.nextElement();
	    String[] tmp2 = (String[]) newTerms.get(tmp);
	    System.out.println("- "+tmp+" \t"+tmp2[0] + tmp2[1]);
	}
    }
    
    public static void main(String[] args){
	
	try{
	    System.out.println("");
	    System.out.println("");
	    System.out.println("");
	    System.out.println("        Terminaattori (versio 0.1). ");    	    
	    System.out.println("");
	    System.out.println("");
	    System.out.println("                  ~~~~_          ");
	    System.out.println("                  ___  \\\\<D      ");              
	    System.out.println("                _/   | /\\\\        ");
	    System.out.println("    _|######## /_____|//%\\\\       ");
	    System.out.println("   / _____   %%%%%%\\__/%%%\\\\___  ");
	    System.out.println("  / /     \\ %%%%%%|      / \\\\   \\ ");              
	    System.out.println("   |       | %%%%/      |   ||   |");
	    System.out.println("   ~~@@@@@@@@~~~/       |        |");
	    System.out.println("    \\_____/              \\______/ ");
	    System.out.println("");
	    System.out.println("");
	    
	    if(args==null || args.length==0 || args[0]==null || args[1]==null) {
		System.out.println("");
		System.out.println("Tarvitsen kutsun yhteydess parametreina: ");
		System.out.println(" 1. polun esinetietoja sisltvn XML-tiedostoon ja");
		System.out.println(" 2. polun asetustiedostoon, jossa on kerrottu:");
		System.out.println("       1. mitk XML-tiedoston kentt luetaan,");
		System.out.println("       2. miss on globaali termikorttitiedosto,");
		System.out.println("       3. miss on lokaali termikorttitiedosto,");
		System.out.println("       4. mihin tiedostoon termit halutaan tulostaa \n"+
				   "          (sek uudet dummyt, ett vanhat lokaalit) ja ");
		System.out.println("       5. mink selitteen haluat termikortin \"Source\"-kenttn \n"+
				   "          (esim. pivmrn ja oman museon nimen)");
		System.out.println("");
	    } else {
		System.out.println("");
		System.out.println("Luetaan olemassa olevat termit.");
		Terminaattori arska = new Terminaattori(args[0], args[1]);
		//arska.testTermsModel();
		System.out.println("");
		System.out.println("Etsitn uudet ilmaisut.");
		arska.findTermCandidatesFromXML(arska.getCards());
		//arska.printNewTerms();
		System.out.println("");
		System.out.println("Kirjoitetaan uudet termikortit:");
		arska.writeNewTermsRDF();
       		System.out.println("");
		System.out.println("\nOlen valmis. Kiitos ajastasi.");
		System.out.println("");
	    }
	}
	catch(Throwable e){
	    e.printStackTrace();
	}
    }
    
}
